from .exceptions import (
    InvalidRestriction,
    LoaderError,
    MissingContextError,
    PyPITokenException,
    ValidationError,
)
from .restrictions import (
    DateRestriction,
    LegacyDateRestriction,
    LegacyNoopRestriction,
    LegacyProjectNamesRestriction,
    ProjectIDsRestriction,
    ProjectNamesRestriction,
    Restriction,
    UserIDRestriction,
)
from .token import Token

__all__ = [
    # Main classes
    "Token",
    "Restriction",
    # Restriction subclasses
    "DateRestriction",
    "ProjectNamesRestriction",
    "ProjectIDsRestriction",
    "UserIDRestriction",
    # Legacy restriction subclasses
    "LegacyNoopRestriction",
    "LegacyProjectNamesRestriction",
    "LegacyDateRestriction",
    # Exceptions
    "PyPITokenException",
    "InvalidRestriction",
    "LoaderError",
    "ValidationError",
    "MissingContextError",
]



from __future__ import annotations

import dataclasses
import datetime
import json
import time
from typing import Any, ClassVar, Iterable, TypeVar

import jsonschema

from pypitoken import exceptions


@dataclasses.dataclass(init=False)
class Context:
    """
    Describe the circumstances sourrounding how the bearer is attempting to use
    the token. Will be checked against restriction to determine whether the bearer
    can or cannot use this token for the operation.

    Parameters
    ----------
    project_name :
        Normalized name of the project the bearer is attempting to upload to
    project_id :
        ID of the project the bearer is attempting to upload to
    user_id :
        ID of the user that is attempting to upload
    now :
        timestamp of the current time, to check date/time-related restrictions.
        Defaults to now.
    """

    # It's expected that no attribute has a default value. Defaults are encoded
    # in __init__ (that's all because the initial values don't match the attribute
    # types)
    project_name: str | None
    project_id: str | None
    user_id: str | None
    now: int

    def __init__(
        self,
        project_name: str | None = None,
        project_id: str | None = None,
        user_id: str | None = None,
        now: int | datetime.datetime | None = None,
    ):
        self.project_name = project_name
        self.project_id = project_id
        self.user_id = user_id

        if now is None:
            self.now = int(time.time())
        else:
            self.now = timestamp_from_parameter(now)


TR = TypeVar("TR", bound="Restriction")


# Making the class a dataclass is mainly meant to ease comparison using ==
@dataclasses.dataclass
class Restriction:
    """
    Base Restriction class.

    Expose lower-level methods for restriction/caveat introspection.
    """

    needs_context: ClassVar[list[str]] = []

    @staticmethod
    def _get_schema() -> dict:
        """
        Return a jsonschema Dict object used to validate the format
        of a json restriction.
        """
        raise NotImplementedError

    @classmethod
    def _load_value(cls: type[TR], value: dict) -> TR:
        """
        Create a Restriction from the JSON value stored in the caveat

        Raises
        ------
        exceptions.LoaderError
            Raise when the JSON format doesn't match this class' restriction format

        Returns
        -------
        Restriction
        """
        schema = cls._get_schema()
        schema["$schema"] = "https://json-schema.org/draft/2020-12/schema"
        try:
            jsonschema.validate(
                instance=value,
                schema=cls._get_schema(),
            )
        except jsonschema.ValidationError as exc:
            raise exceptions.LoaderError() from exc

        return cls(**cls._extract_kwargs(value=value))  # type: ignore

    @staticmethod
    def _get_subclasses() -> list[type[Restriction]]:
        """
        List all subclasses of Restriction that we want to match against
        """
        # We could use __subclasses__ but that could lead to all kinds of funky things,
        # especially in a security-sensistive library.
        # Tests will check this against Restriction subclasses though.
        return [
            DateRestriction,
            ProjectNamesRestriction,
            ProjectIDsRestriction,
            UserIDRestriction,
            # Legacy
            LegacyNoopRestriction,
            LegacyProjectNamesRestriction,
            LegacyDateRestriction,
        ]

    @staticmethod
    def _json_load_caveat(caveat: str) -> Any:
        try:
            value = json.loads(caveat)
        except Exception as exc:
            raise exceptions.LoaderError(f"Error while loading caveat: {exc}") from exc

        return value

    @classmethod
    def load(cls, caveat: dict) -> Restriction:
        """
        Create a Restriction from a raw caveat restriction JSON object.

        Raises
        ------
        pypitokens.LoaderError
            If the format cannot be understood

        Returns
        -------
        `Restriction`
        """
        for subclass in cls._get_subclasses():
            try:
                return subclass._load_value(value=caveat)
            except exceptions.LoaderError:
                continue

        raise exceptions.LoaderError(
            f"Could not find matching Restriction for {caveat}"
        )

    @classmethod
    def load_json(cls, caveat: str) -> Restriction:
        """
        Create a Restriction from a raw caveat restriction JSON string.

        Raises
        ------
        pypitokens.LoaderError
            If the format cannot be understood

        Returns
        -------
        `Restriction`
        """
        caveat_obj = cls._json_load_caveat(caveat=caveat)
        return cls.load(caveat=caveat_obj)

    @classmethod
    def _extract_kwargs(cls, value: Any) -> dict:
        """
        Receive the parsed JSON value of a caveat for which the schema has been
        validated. Return the instantiation kwargs (``__init__`` parameters).
        """
        raise NotImplementedError

    @classmethod
    def from_parameters(cls: type[TR], **kwargs) -> TR | None:
        """
        Constructs an instance from the parameters passed to `Token.restrict`
        """
        raise NotImplementedError

    @classmethod
    def restrictions_from_parameters(cls, **kwargs) -> Iterable[Restriction]:
        """
        Constructs an iterable of Restriction subclass instances from the parameters
        passed to `Token.restrict`
        """
        for subclass in cls._get_subclasses():
            restriction = subclass.from_parameters(**kwargs)
            if restriction:
                yield restriction

    def check(self, context: Context) -> None:
        """
        Receive the context of a check

        Parameters
        ----------
        context :
            Describes how the bearer is attempting to use the token.

        Raises
        ------
        `ValidationError`
            Restriction was checked and appeared unmet.
        `MissingContextError`
            (a subclass of ValidationError): Restriction is unmet because the
            context is lacking required values. This is more probably a code issue
            from the `Token.check` call than an issue with the token itself,
            though it also means that the token should be rejected.
        """
        for need in self.needs_context:
            if getattr(context, need) is None:
                raise exceptions.MissingContextError(
                    "The restriction couldn't be checked because the context is missing required values"
                )

    def dump(self) -> dict | list:
        """
        Transform a restriction into a JSON-compatible dict object
        """
        raise NotImplementedError

    def dump_json(self) -> str:
        """
        Transform a restriction into a JSON-encoded string
        """
        return json.dumps(self.dump())


@dataclasses.dataclass
class DateRestriction(Restriction):
    """
    Restrict a `Token` to a single time interval.

    Attributes
    ----------
    not_before :
        Token is not to be used before this Unix timestamp.
    not_after :
        Token is not to be used after this Unix timestamp.
    """

    not_before: int
    not_after: int
    # https://github.com/pypi/warehouse/blob/606bce63472a7979eff0a37ea8c44c335b9c118f/warehouse/macaroons/caveats/__init__.py#L41
    tag: ClassVar[int] = 0
    needs_context: ClassVar[list[str]] = ["now"]

    @classmethod
    def _get_schema(cls) -> dict:
        return {
            "type": "array",
            "prefixItems": [
                {"type": "integer", "const": cls.tag},
                {"type": "integer"},
                {"type": "integer"},
            ],
            "items": False,
            "minItems": 3,
            "maxItems": 3,
        }

    @classmethod
    def _extract_kwargs(cls, value: list) -> dict:
        return {
            "not_before": value[1],
            "not_after": value[2],
        }

    def dump(self) -> list:
        return [self.tag, self.not_before, self.not_after]

    def check(self, context: Context) -> None:
        super().check(context=context)
        now: int = context.now

        if not self.not_before <= now < self.not_after:
            raise exceptions.ValidationError(
                f"This token can only be used between timestamps "
                f"{self.not_before} (incl) and {self.not_after} (excl). "
                f"Received: {now}"
            )

    @classmethod
    def from_parameters(
        cls,
        not_before: int | datetime.datetime | None = None,
        not_after: int | datetime.datetime | None = None,
        **kwargs,
    ) -> DateRestriction | None:
        if not (not_before or not_after):
            return None

        if not (not_before and not_after):
            raise exceptions.InvalidRestriction(
                "`not_before` and `not_after` parameters must be used together. "
                "Either define both or neither. "
                f"Received not_before={not_before} and not_after={not_after}"
            )
        return cls(
            not_before=timestamp_from_parameter(not_before),
            not_after=timestamp_from_parameter(not_after),
        )


@dataclasses.dataclass
class ProjectNamesRestriction(Restriction):
    """
    Restrict a `Token` to uploading releases for project with specific names.

    Attributes
    ----------
    project_names :
        Normalized project names this token may upload to.
    """

    project_names: list[str]
    # https://github.com/pypi/warehouse/blob/606bce63472a7979eff0a37ea8c44c335b9c118f/warehouse/macaroons/caveats/__init__.py#L54
    tag: ClassVar[int] = 1
    needs_context: ClassVar[list[str]] = ["project_name"]

    @classmethod
    def _get_schema(cls) -> dict:
        return {
            "type": "array",
            "prefixItems": [
                {"type": "integer", "const": cls.tag},
                {
                    "type": "array",
                    "items": {
                        "type": "string",
                        # Regex from https://peps.python.org/pep-0426/#name
                        # with normalization rule from
                        # https://peps.python.org/pep-0503/#normalized-names
                        "pattern": r"^([a-z0-9]|[a-z0-9][a-z0-9-]*[a-z0-9])$",
                    },
                },
            ],
            "minItems": 2,
            "maxItems": 2,
            "items": False,
        }

    @classmethod
    def _extract_kwargs(cls, value: list) -> dict:
        return {"project_names": value[1]}

    def dump(self) -> list:
        return [self.tag, self.project_names]

    def check(self, context: Context) -> None:
        super().check(context=context)
        project_name = context.project_name

        if project_name not in self.project_names:
            raise exceptions.ValidationError(
                f"This token can only be used for project(s): "
                f"{', '.join(self.project_names)}. Received: {project_name}"
            )

    @classmethod
    def from_parameters(
        cls,
        project_names: list[str] | None = None,
        **kwargs,
    ) -> ProjectNamesRestriction | None:
        if project_names is not None:
            return cls(project_names=project_names)
        return None


@dataclasses.dataclass
class ProjectIDsRestriction(Restriction):
    """
    Restrict a `Token` to uploading releases for project with specific IDs.

    Attributes
    ----------
    project_ids :
        Project IDs this token may upload to.
    """

    project_ids: list[str]
    # https://github.com/pypi/warehouse/blob/606bce63472a7979eff0a37ea8c44c335b9c118f/warehouse/macaroons/caveats/__init__.py#L71
    tag: ClassVar[int] = 2
    needs_context: ClassVar[list[str]] = ["project_id"]

    @classmethod
    def _get_schema(cls) -> dict:
        return {
            "type": "array",
            "prefixItems": [
                {"type": "integer", "const": cls.tag},
                {
                    "type": "array",
                    "items": {
                        "type": "string",
                        # Uppercase UUIDs are not accepted
                        "pattern": r"^[0-9a-f]{8}\b-[0-9a-f]{4}\b-[0-9a-f]{4}\b-[0-9a-f]{4}\b-[0-9a-f]{12}$",
                    },
                },
            ],
            "minItems": 2,
            "maxItems": 2,
            "items": False,
        }

    @classmethod
    def _extract_kwargs(cls, value: list) -> dict:
        return {"project_ids": value[1]}

    def dump(self) -> list:
        return [self.tag, self.project_ids]

    def check(self, context: Context) -> None:
        super().check(context=context)
        project_id = context.project_id

        if project_id not in self.project_ids:
            raise exceptions.ValidationError(
                f"This token can only be used for project(s): "
                f"{', '.join(self.project_ids)}. Received: {project_id}"
            )

    @classmethod
    def from_parameters(
        cls,
        project_ids: list[str] | None = None,
        **kwargs,
    ) -> ProjectIDsRestriction | None:
        if project_ids is not None:
            return cls(project_ids=project_ids)
        return None


@dataclasses.dataclass
class UserIDRestriction(Restriction):
    """
    Restrict a `Token` to being used by the user with the corresponding ID.

    Attributes
    ----------
    user_id :
        ID of the user that may use this token.
    """

    user_id: str
    # https://github.com/pypi/warehouse/blob/606bce63472a7979eff0a37ea8c44c335b9c118f/warehouse/macaroons/caveats/__init__.py#L88
    tag: ClassVar[int] = 3
    needs_context: ClassVar[list[str]] = ["user_id"]

    @classmethod
    def _get_schema(cls) -> dict:
        return {
            "type": "array",
            "prefixItems": [
                {"type": "integer", "const": cls.tag},
                {
                    "type": "string",
                    # Uppercase ID is not accepted
                    "pattern": r"^[0-9a-f]{8}\b-[0-9a-f]{4}\b-[0-9a-f]{4}\b-[0-9a-f]{4}\b-[0-9a-f]{12}$",
                },
            ],
            "minItems": 2,
            "maxItems": 2,
            "items": False,
        }

    @classmethod
    def _extract_kwargs(cls, value: list) -> dict:
        return {"user_id": value[1]}

    def dump(self) -> list:
        return [self.tag, self.user_id]

    def check(self, context: Context) -> None:
        super().check(context=context)
        user_id = context.user_id

        if user_id != self.user_id:
            raise exceptions.ValidationError(
                f"This token can only be used by user with id: "
                f"{self.user_id}. Received: {user_id}"
            )

    @classmethod
    def from_parameters(
        cls,
        user_id: str | None = None,
        **kwargs,
    ) -> UserIDRestriction | None:
        if user_id is not None:
            return cls(user_id=user_id)
        return None


@dataclasses.dataclass
class LegacyNoopRestriction(Restriction):
    """
    Says it restricts the `Token`, but doesn't actually restrict it.
    """

    @staticmethod
    def _get_schema() -> dict:
        return {
            "type": "object",
            "properties": {
                "version": {"type": "integer", "const": 1},
                "permissions": {"type": "string", "const": "user"},
            },
            "required": ["version", "permissions"],
            "additionalProperties": False,
        }

    @classmethod
    def _extract_kwargs(cls, value: dict) -> dict:
        return {}

    def check(self, context: Context) -> None:
        super().check(context=context)
        # Always passes
        return

    def dump(self) -> dict:
        return {"version": 1, "permissions": "user"}

    @classmethod
    def from_parameters(
        cls,
        legacy_noop: bool | None = None,
        **kwargs,
    ) -> LegacyNoopRestriction | None:
        if legacy_noop is True:
            return cls()
        return None


@dataclasses.dataclass
class LegacyProjectNamesRestriction(Restriction):
    """
    Restrict a `Token` to uploading releases for a specific set of packages.

    Attributes
    ----------
    projects :
        Normalized project names this token may upload to.
    """

    project_names: list[str]
    needs_context: ClassVar[list[str]] = ["project_name"]

    @staticmethod
    def _get_schema() -> dict:
        return {
            "type": "object",
            "properties": {
                "version": {"type": "integer", "const": 1},
                "permissions": {
                    "type": "object",
                    "properties": {
                        "projects": {"type": "array", "items": {"type": "string"}}
                    },
                    "required": ["projects"],
                    "additionalProperties": False,
                },
            },
            "required": ["version", "permissions"],
            "additionalProperties": False,
        }

    @classmethod
    def _extract_kwargs(cls, value: dict) -> dict:
        return {"project_names": value["permissions"]["projects"]}

    def dump(self) -> dict:
        return {"version": 1, "permissions": {"projects": self.project_names}}

    def check(self, context: Context) -> None:
        super().check(context=context)
        project_name = context.project_name

        if project_name not in self.project_names:
            raise exceptions.ValidationError(
                f"This token can only be used for project(s): "
                f"{', '.join(self.project_names)}. Received: {project_name}"
            )

    @classmethod
    def from_parameters(
        cls,
        legacy_project_names: list[str] | None = None,
        **kwargs,
    ) -> LegacyProjectNamesRestriction | None:
        if legacy_project_names is not None:
            return cls(project_names=legacy_project_names)
        return None


@dataclasses.dataclass
class LegacyDateRestriction(Restriction):
    """
    Restrict a `Token` to a single time interval.

    Attributes
    ----------
    not_before :
        Token is not to be used before this Unix timestamp.
    not_after :
        Token is not to be used after this Unix timestamp.

    """

    not_before: int
    not_after: int
    needs_context: ClassVar[list[str]] = ["now"]

    @staticmethod
    def _get_schema() -> dict:
        return {
            "type": "object",
            "properties": {
                "nbf": {"type": "integer"},
                "exp": {"type": "integer"},
            },
            "required": ["nbf", "exp"],
            "additionalProperties": False,
        }

    @classmethod
    def _extract_kwargs(cls, value: dict) -> dict:
        return {
            "not_before": value["nbf"],
            "not_after": value["exp"],
        }

    def dump(self) -> dict:
        return {"nbf": self.not_before, "exp": self.not_after}

    def check(self, context: Context) -> None:
        super().check(context=context)
        now: int = context.now

        if not self.not_before <= now < self.not_after:
            raise exceptions.ValidationError(
                f"This token can only be used between timestamps "
                f"{self.not_before} (incl) and {self.not_after} (excl). "
                f"Received: {now}"
            )

    @classmethod
    def from_parameters(
        cls,
        legacy_not_before: int | datetime.datetime | None = None,
        legacy_not_after: int | datetime.datetime | None = None,
        **kwargs,
    ) -> LegacyDateRestriction | None:
        if legacy_not_before or legacy_not_after:
            if not (legacy_not_before and legacy_not_after):
                raise exceptions.InvalidRestriction(
                    "`legacy_not_before` and `legacy_not_after` parameters "
                    "must be used together. Either define both or neither. "
                    f"Received legacy_not_before={legacy_not_before} and "
                    f"legacy_not_after={legacy_not_after}"
                )
            return cls(
                not_before=timestamp_from_parameter(legacy_not_before),
                not_after=timestamp_from_parameter(legacy_not_after),
            )
        return None


def timestamp_from_parameter(param: int | datetime.datetime) -> int:
    if isinstance(param, int):
        return int(param)
    # https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
    naive = param.tzinfo is None or param.tzinfo.utcoffset(param) is None
    if naive:
        raise exceptions.InvalidRestriction(
            "Cannot use a naive datetime. Either provide a timezone or "
            "the timestamp directly. "
            "Received {param}"
        )
    return int(param.timestamp())



from __future__ import annotations


class PyPITokenException(Exception):
    """
    The base exception for all exceptions raised
    from this library. Any code raising a different
    exception would be considered a bug.
    """


class InvalidRestriction(PyPITokenException, ValueError):
    """
    Exception encountered while calling `Token.restrict`, due to unexpected
    parameters.
    """


class LoaderError(PyPITokenException):
    """
    Exception encountered while calling `Token.load`, due to unexpected
    format.

    Exception should be associated with a message in English that can be shown to the
    bearer to explain the error.
    """


class ValidationError(PyPITokenException):
    """
    Exception encountered while calling `Token.check`, the token should
    be considered invalid.

    Exception should be associated with a message in English that can be shown to the
    bearer to explain the error.
    """


class MissingContextError(ValidationError):
    """
    Exception encountered while calling `Token.check`, the token should
    be considered invalid.

    The restriction couldn't be checked because the context is missing required
    values
    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypitoken']

package_data = \
{'': ['*']}

install_requires = \
['jsonschema', 'pymacaroons>=0.13.0,<0.14.0', 'typing-extensions']

setup_kwargs = {
    'name': 'pypitoken',
    'version': '6.0.0',
    'description': 'Manipulate PyPI API tokens',
    'long_description': 'PyPIToken: Manipulate PyPI API tokens\n=====================================\n\n.. image:: https://img.shields.io/pypi/v/pypitoken?logo=pypi&logoColor=white\n    :target: https://pypi.org/pypi/pypitoken\n    :alt: Deployed to PyPI\n\n.. image:: https://img.shields.io/pypi/pyversions/pypitoken?logo=pypi&logoColor=white\n    :target: https://pypi.org/pypi/pypitoken\n    :alt: Deployed to PyPI\n\n.. image:: https://img.shields.io/github/stars/ewjoachim/pypitoken?logo=github\n    :target: https://github.com/ewjoachim/pypitoken/\n    :alt: GitHub Repository\n\n.. image:: https://img.shields.io/github/workflow/status/ewjoachim/pypitoken/CI?logo=github\n    :target: https://github.com/ewjoachim/pypitoken/actions?workflow=CI\n    :alt: Continuous Integration\n\n.. image:: https://img.shields.io/readthedocs/pypitoken?logo=read-the-docs&logoColor=white\n    :target: http://pypitoken.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation\n\n.. image:: https://img.shields.io/endpoint?logo=codecov&logoColor=white&url=https://raw.githubusercontent.com/wiki/ewjoachim/pypitoken/coverage-comment-badge.json\n    :target: https://github.com/marketplace/actions/coverage-comment\n    :alt: Coverage\n\n.. image:: https://img.shields.io/github/license/ewjoachim/pypitoken?logo=open-source-initiative&logoColor=white\n    :target: https://github.com/ewjoachim/pypitoken/blob/master/LICENSE\n    :alt: MIT License\n\n.. image:: https://img.shields.io/badge/Contributor%20Covenant-v1.4%20adopted-ff69b4.svg\n    :target: https://github.com/ewjoachim/pypitoken/blob/master/CODE_OF_CONDUCT.md\n    :alt: Contributor Covenant\n\n\nPyPIToken is an open-source Python 3.7+ library for generating and manipulating\nPyPI tokens.\n\nPyPI tokens are very powerful, as that they are based on Macaroons_. They allow\nthe bearer to add additional restrictions to an existing token. For example, given\na PyPI token that can upload releases for any project of its owner, you can generate\na token that will only allow some projects, or even a single one.\n\n.. _macaroons: https://en.wikipedia.org/wiki/Macaroons_(computer_science)\n\nHere\'s an example:\n\n.. code-block:: console\n\n    $ pip install pypitoken\n\n.. code-block:: python\n\n    import pypitoken\n\n    token = pypitoken.Token.load("pypi-foobartoken")\n\n    print(token.restrictions)\n    # [ProjectIDsRestriction(project_ids=["00000000-0000-0000-0000-000000000000"])]\n\n    token.restrict(project_names=["requests"])\n\n    print(token.restrictions)\n    # [\n    #     ProjectIDsRestriction(project_ids=["00000000-0000-0000-0000-000000000000"]),\n    #     ProjectNamesRestriction(project_names=["requests"]),\n    # ]\n\n    token.dump()\n    # pypi-newfoobartoken\n\nThis token we\'ve created above will be restricted to uploading releases of ``requests``.\nOf course, your PyPI user will still need to have upload permissions on ``requests``\nfor this to happen.\n\nThe aim of this library is to provide a simple toolbelt for manipulating PyPI tokens.\nIdeally, someday, PyPI (Warehouse_) itself may generate their tokens using this\nlibrary too. This should make it easier to iterate on new kinds of restrictions for\nPyPI tokens, such as those discussed in the `original implementation issue`__.\n\n.. _Warehouse: https://github.com/pypa/warehouse/\n.. __: https://github.com/pypa/warehouse/issues/994\n\nA discussion for integrating this library to the Warehouse environment is ongoing:\n\n- In the `Python Packaging discussions`_ for putting the project under the PyPA umbrella\n- In the `Warehouse tracker`_ for replacing the current macaroon implementation with\n  this lib\n\n.. _`Python Packaging discussions`: https://discuss.python.org/t/pypitoken-a-library-for-generating-and-manipulating-pypi-tokens/7572\n.. _`Warehouse tracker`: https://github.com/pypa/warehouse/issues/9184\n\n.. Below this line is content specific to the README that will not appear in the doc.\n.. end-of-index-doc\n\nWhere to go from here\n---------------------\n\nThe complete docs_ is probably the best place to learn about the project.\n\nIf you encounter a bug, or want to get in touch, you\'re always welcome to open a\nticket_.\n\n.. _docs: http://pypitoken.readthedocs.io/en/latest\n.. _ticket: https://github.com/ewjoachim/pypitoken/issues/new\n',
    'author': 'Joachim Jablon',
    'author_email': 'ewjoachim@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pypitoken.readthedocs.io/en/latest/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import datetime
import functools

import pymacaroons
from typing_extensions import ParamSpec

from pypitoken import exceptions, restrictions

PREFIX = "pypi"

P = ParamSpec("P")


class Token:
    """
    Create a `Token` (as PyPI itself would do), load an existing
    `Token`, create additional restrictions on a `Token` and dump a `Token` string
    from a `Token`.

    This class is an higher level abstraction over the Macaroons from pymacaroons,
    with specific knowledge of their implementation within PyPI.

    Attributes
    ----------
    prefix : str
        PyPI tokens are usually prefixed with ``pypi``, but this is arbitrary
    domain : str
        PyPI tokens are attached to a specific domain, usually ``pypi.org`` or
        ``test.pypi.org``
    identifier : str
        This part is how PyPI will find a token in its database and associate it to a
        specific user. Even when additional restrictions are added to a token, the
        identifier will still be readable in the token. While this is not exactly
        a secret part of the token, it's best to keep it reasonably private.
    _macaroon : pymacarrons.Macaroon
        This is part of the private API and may be subject to change at any time.
        This gives you access to the underlying Macaroon, if you need to do low-level
        operations, or just want to poke around.
    """

    def __init__(self, prefix: str, macaroon: pymacaroons.Macaroon):
        """
        This constructor is usually not meant to be called directly, use either
        `create` or `load` to get `Token` instances.

        Parameters
        ----------
        prefix :
            prefix of the token, usually `pypi`
        macaroon :
            Macaroon containing all the token data
        """
        self.prefix = prefix
        self._macaroon = macaroon

    @property
    def domain(self) -> str:
        return self._macaroon.location

    @property
    def identifier(self) -> str:
        return self._macaroon.identifier.decode("utf-8")

    @classmethod
    def load(cls, raw: str) -> Token:
        """
        Deserialize a `Token` from a raw string.
        Warning: does NOT check for the Token validity, you need to call `check`
        for that.

        Parameters
        ----------
        raw :
            The token string (including the prefix)

        Returns
        -------
        Token
            Loaded token

        Raises
        ------
        `pypitoken.LoaderError`
            Any error in loading the token will be raised as a LoaderError.
            The original exception (if any) will be attached
            as the exception cause (``raise from``).
        """
        try:
            prefix, raw_macaroon = raw.split("-", maxsplit=1)
        except ValueError:
            raise exceptions.LoaderError("Token is missing a prefix")
        try:
            macaroon = pymacaroons.Macaroon.deserialize(raw_macaroon)
        # https://github.com/ecordell/pymacaroons/issues/50
        except Exception as exc:
            raise exceptions.LoaderError(f"Deserialization error: {exc}") from exc
        return cls(
            prefix=prefix,
            macaroon=macaroon,
        )

    @classmethod
    def create(
        cls,
        domain: str,
        identifier: str,
        key: str | bytes,
        prefix: str = PREFIX,
        version: int = pymacaroons.MACAROON_V2,
    ) -> Token:
        """
        Create a token. Initially, it has no restruction, but they can be added
        with restrict.
        In order to create the token, you will need to provide a key. This key is the
        secret part of the token. When checking a macaroon validity, you will need to
        provide the same key.

        Parameters
        ----------
        domain :
            PyPI tokens are attached to a specific domain, usually ``pypi.org`` or
            ``test.pypi.org``.
        identifier :
            Identifies the Token (and its associated user) in the PyPI database.
        key :
            Secret key used to validate the token. Having the key of a token would
            allow removing existing restrictions.
        prefix :
            PyPI tokens are usually prefixed with ``pypi`` which is the default value
            for this parameter.
        version :
            Version of the pymacaroons specification to use. There's probably no reason
            we would want to change this.

        Returns
        -------
        `Token`
            The newly minted token
        """
        macaroon = pymacaroons.Macaroon(
            location=domain,
            identifier=identifier,
            key=key,
            version=version,
        )
        token = cls(prefix=prefix, macaroon=macaroon)
        return token

    def restrict(
        self,
        not_before: int | datetime.datetime | None = None,
        not_after: int | datetime.datetime | None = None,
        project_names: list[str] | None = None,
        project_ids: list[str] | None = None,
        user_id: str | None = None,
        # Legacy params
        legacy_project_names: list[str] | None = None,
        legacy_not_before: int | datetime.datetime | None = None,
        legacy_not_after: int | datetime.datetime | None = None,
        legacy_noop: bool | None = None,
    ) -> Token:
        """
        Modifies the token in-place to add restrictions to it. This can be called by
        PyPI as well as by anyone, adding restrictions to new or existing tokens.

        Multiple restrictions of different types can be added in one call to
        `restrict()`. Alternatively, multiple restrictions of the same type can be added
        via multiple calls to `restrict()`.

        Note: a token allows the owner to delegate their rights to the bearer.
        Consequently, a token adding restrictions linked to a project that the owner
        cannot use will not make its bearer able to upload releases to the project. As
        stated in the name, restrictions can reduce the scope of a token, and cannot
        broaden it. It's possible to create a token that can do nothing, for example by
        adding two incompatible restrictions.

        Parameters
        ----------

        not_before :
            Restrict the token to uploading releases only after the given timestamp or
            tz-aware datetime. Must be used with ``not_after``. By default, None (no
            restriction)
        not_after :
            Restrict the token to uploading releases only before the given timestamp or
            tz-aware datetime. Must be used with ``not_before``. By default, None (no
            restriction)
        project_names :
            Restrict the token to uploading releases only for projects with these names,
            by default None (no restriction)
        project_ids :
            Restrict the token to uploading releases only for projects with these ids,
            by default None (no restriction)
        user_ids :
            Restrict the token to uploading user attempting to upload Restrict the token
            to being used only by user with this id, by default None (no restriction)
        legacy_not_before :
            Uses the legacy restriction format (results in longer token size). Restrict
            the token to uploading releases only after the given timestamp or tz-aware
            datetime. Must be used with ``not_after``. By default, None (no restriction)
        legacy_not_after :
            Uses the legacy restriction format (results in longer token size). Restrict
            the token to uploading releases only before the given timestamp or tz-aware
            datetime. Must be used with ``not_before``. By default, None (no
            restriction)
        legacy_project_names :
            Uses the legacy restriction format (results in longer token size). Restrict
            the token to uploading releases only for projects with these ed name of the
            project the bearer is attempting to upload toed names, by default None (no
            restriction)
        legacy_noop :
            Uses the legacy restriction format (results in longer token size). user
            attempting to upload This restriction being there or not doesn't change the
            validity of the token.

        Raises
        ------
        exceptions.Invalidrestrictions.Restriction
            If the provided parameters cannot describe a valid description

        Returns
        -------
        `Token`
            The modified Token, to ease chaining calls.
        """
        for restriction in restrictions.Restriction.restrictions_from_parameters(
            not_before=not_before,
            not_after=not_after,
            project_names=project_names,
            project_ids=project_ids,
            user_id=user_id,
            legacy_project_names=legacy_project_names,
            legacy_not_before=legacy_not_before,
            legacy_not_after=legacy_not_after,
            legacy_noop=legacy_noop,
        ):
            self._macaroon.add_first_party_caveat(restriction.dump_json())

        return self

    def dump(self) -> str:
        """
        Generates a string representing the token. This could be used for
        API authentication against PyPI.

        Returns
        -------
        str
            The token
        """
        return f"{self.prefix}-{self._macaroon.serialize()}"

    def check(
        self,
        key: str | bytes,
        project_name: str | None = None,
        project_id: str | None = None,
        user_id: str | None = None,
        now: int | datetime.datetime | None = None,
    ) -> None:
        """
        Raises pypitoken.ValidationError if the token is invalid.

        Parameters besides ``key`` are all optional and will be used to provide the
        current context this token is used for, allowing to accept or reject the caveat
        restrictions. If a parameter is not passed, but a caveat using it is
        encountered, the caveat will not be met, the token will be found invalid, and
        this method will raise with an appropriate exception. There is an exception for
        ``now``: if not passed, it defaults to the current timestamp.

        Parameters
        ----------
        key : str
            Key of the macaroon, stored in PyPI database
        project_name :
            Normalized name of the project the bearer is attempting to upload to
        project_id :
            ID of the project the bearer is attempting to upload to
        user_id :
            ID of the user attempting to upload
        now :
            Timestamp of the moment the upload takes place, or tz-aware datetime.
            Defaults to the current timestamp.

        Raises
        ------
        `pypitoken.ValidationError`
            Any error in validating the token will be raised as a ValidationError. The
            original exception (if any) will be attached as the exception cause (``raise
            from``).
        """
        verifier = pymacaroons.Verifier()

        context = restrictions.Context(
            project_name=project_name,
            project_id=project_id,
            user_id=user_id,
            now=now,
        )

        errors: list[Exception] = []

        verifier.satisfy_general(
            functools.partial(self._check_caveat, context=context, errors=errors)
        )
        try:
            verifier.verify(self._macaroon, key)
        # https://github.com/ecordell/pymacaroons/issues/51
        except Exception as exc:
            if errors:
                # (we know it's actually a single item, there cannot be multiple items
                # in this list)
                (exc,) = errors

            raise exceptions.ValidationError(
                f"Error while validating token: {exc}"
            ) from exc

    @staticmethod
    def _check_caveat(
        caveat: str,
        context: restrictions.Context,
        errors: list[Exception],
    ) -> bool:
        """
        This method follows the pymacaroon Verifier.satisfy_general API, except
        that it takes a context parameter. It's expected to be used with
        `functools.partial`

        Parameters
        ----------
        caveat :
            raw caveat string that will be turned into a restriction to be checked
        context :
            Dict describing what we're trying to use the Token for, in order to decide
            whether the restrictions expressed by the caveat apply.
        errors :
            If any exception is raised, it will be appended to this list, so that
            we'll be able to raise them properly later

        Returns
        -------
        bool
            True if the caveat is met, False otherwise (should not raise).
        """

        try:
            restriction = restrictions.Restriction.load_json(caveat=caveat)
        except exceptions.LoaderError as exc:
            errors.append(exc)
            return False

        try:
            restriction.check(context=context)
        except exceptions.ValidationError as exc:
            errors.append(exc)
            return False

        return True

    @property
    def restrictions(self) -> list[restrictions.Restriction]:
        """
        Return a list of restrictions associated to this `Token`. This can be used
        to get a better insight on what this `Token` contains.

        Returns
        -------
        List[`restrictions.Restriction`]

        Raises
        ------
        `pypitoken.LoaderError`
            When the existing restrictions cannot be parsed
        """
        return [
            restrictions.Restriction.load_json(caveat=caveat.caveat_id)
            for caveat in self._macaroon.caveats
        ]
